import pytest
from anyio import TASK_STATUS_IGNORED, create_memory_object_stream, create_task_group, sleep
from anyio.abc import TaskStatus
from pycrdt import (
    Array,
    Doc,
    create_sync_message,
    create_update_message,
    handle_sync_message,
)
from pycrdt._sync import Decoder, write_var_uint

pytestmark = pytest.mark.anyio


class ConnectedDoc:
    def __init__(self):
        self.doc = Doc()
        self.doc.observe(lambda event: self.send(event.update))
        self.connected_docs = []
        self.send_stream, self.receive_stream = create_memory_object_stream[bytes](
            max_buffer_size=1024
        )

    async def start(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED):
        async with create_task_group() as tg:
            task_status.started()
            tg.start_soon(self.process_received_messages)

    def connect(self, *connected_docs):
        self.connected_docs += connected_docs
        sync_message = create_sync_message(self.doc)
        for connected_doc in connected_docs:
            connected_doc.receive(sync_message, self)

    def receive(self, message: bytes, sender=None):
        self.send_stream.send_nowait((message, sender))

    async def process_received_messages(self):
        async for message, sender in self.receive_stream:
            reply = handle_sync_message(message[1:], self.doc)
            if reply is not None:
                sender.receive(reply, self)

    def send(self, message: bytes):
        for doc in self.connected_docs:
            doc.receive(create_update_message(message))


async def test_sync():
    async with create_task_group() as tg:
        doc0 = ConnectedDoc()
        doc1 = ConnectedDoc()

        await tg.start(doc0.start)
        await tg.start(doc1.start)

        doc0.connect(doc1)
        doc1.connect(doc0)

        array0 = doc0.doc.get("array", type=Array)
        array0.append(0)
        await sleep(0.1)

        array1 = doc1.doc.get("array", type=Array)
        assert array1[0] == 0

        # doc2 only connects to doc0
        # but since doc0 and doc1 are connected,
        # doc2 is indirectly connected to doc1
        doc2 = ConnectedDoc()
        await tg.start(doc2.start)
        doc2.connect(doc0)
        doc0.connect(doc2)
        await sleep(0.1)
        array2 = doc2.doc.get("array", type=Array)
        assert array2[0] == 0
        array2.append(1)
        await sleep(0.1)
        assert array0[1] == 1
        assert array1[1] == 1

        tg.cancel_scope.cancel()


def test_write_var_uint():
    assert write_var_uint(128) == b"\x80\x01"


def test_decoder():
    with pytest.raises(RuntimeError) as exc_info:
        Decoder(b"").read_var_uint()
    assert str(exc_info.value) == "Y protocol error"

    assert list(Decoder(b"").read_messages()) == []
    assert list(Decoder(b"\x00").read_messages()) == [b""]
    assert Decoder(b"").read_var_string() == ""
    assert Decoder(b"\x05Hello").read_var_string() == "Hello"



from __future__ import annotations

from typing import Callable, Type, TypeVar, cast

from ._base import BaseDoc, BaseType, base_types
from ._pycrdt import Doc as _Doc
from ._pycrdt import SubdocsEvent, Subscription, TransactionEvent
from ._pycrdt import Transaction as _Transaction
from ._transaction import ReadTransaction, Transaction

T_BaseType = TypeVar("T_BaseType", bound=BaseType)


class Doc(BaseDoc):
    def __init__(
        self,
        init: dict[str, BaseType] = {},
        *,
        client_id: int | None = None,
        doc: _Doc | None = None,
        Model=None,
    ) -> None:
        super().__init__(client_id=client_id, doc=doc, Model=Model)
        for k, v in init.items():
            self[k] = v
        if Model is not None:
            self._twin_doc = Doc(init)

    @property
    def guid(self) -> int:
        return self._doc.guid()

    @property
    def client_id(self) -> int:
        return self._doc.client_id()

    def transaction(self) -> Transaction:
        if self._txn is not None:
            return self._txn
        return Transaction(self)

    def _read_transaction(self, _txn: _Transaction) -> ReadTransaction:
        return ReadTransaction(self, _txn)

    def get_state(self) -> bytes:
        return self._doc.get_state()

    def get_update(self, state: bytes | None = None) -> bytes:
        if state is None:
            state = b"\x00"
        return self._doc.get_update(state)

    def apply_update(self, update: bytes) -> None:
        if self._Model is not None:
            twin_doc = cast(Doc, self._twin_doc)
            twin_doc.apply_update(update)
            d = {k: twin_doc[k].to_py() for k in self._Model.model_fields}
            try:
                self._Model(**d)
            except Exception as e:
                self._twin_doc = Doc(dict(self))
                raise e
        self._doc.apply_update(update)

    def __setitem__(self, key: str, value: BaseType) -> None:
        if not isinstance(key, str):
            raise RuntimeError("Key must be of type string")
        integrated = value._get_or_insert(key, self)
        prelim = value._integrate(self, integrated)
        value._init(prelim)

    def __getitem__(self, key: str) -> BaseType:
        return self._roots[key]

    def __iter__(self):
        return iter(self.keys())

    def get(self, key: str, *, type: type[T_BaseType]) -> T_BaseType:
        value = type()
        self[key] = value
        return value

    def keys(self):
        return self._roots.keys()

    def values(self):
        return self._roots.values()

    def items(self):
        return self._roots.items()

    @property
    def _roots(self) -> dict[str, BaseType]:
        with self.transaction() as txn:
            assert txn._txn is not None
            return {
                key: (
                    None
                    if val is None
                    else cast(Type[BaseType], base_types[type(val)])(_integrated=val, _doc=self)
                )
                for key, val in self._doc.roots(txn._txn).items()
            }

    def observe(self, callback: Callable[[TransactionEvent], None]) -> Subscription:
        subscription = self._doc.observe(callback)
        self._subscriptions.append(subscription)
        return subscription

    def observe_subdocs(self, callback: Callable[[SubdocsEvent], None]) -> Subscription:
        subscription = self._doc.observe_subdocs(callback)
        self._subscriptions.append(subscription)
        return subscription

    def unobserve(self, subscription: Subscription) -> None:
        self._subscriptions.remove(subscription)
        subscription.drop()


base_types[_Doc] = Doc



import json
from functools import partial

import pytest
from pycrdt import Array, Doc, Map, Text


def callback(events, event):
    events.append(event)


def callback_deep(_events, events):
    _events.append(events)


def test_binary_entry():
    doc = Doc({"m": Map({"bytes": b"012"})})
    assert doc["m"]["bytes"] == b"012"


def test_str():
    doc = Doc()
    map2 = Map({"key2": "val2"})
    array1 = Array([0, 1, map2])
    map0 = Map({"key1": array1})
    doc["map"] = map0
    assert str(map0) == '{"key1":[0,1,{"key2":"val2"}]}'


def test_nested():
    doc = Doc()
    text1 = Text("my_text")
    array1 = Array([0, "foo", 2])
    map1 = Map({"foo": [3, 4, 5], "bar": "baz"})
    map0 = Map({"text1": text1, "array1": array1, "map1": map1})
    doc["map"] = map0
    ref = {
        "text1": "my_text",
        "array1": [0, "foo", 2],
        "map1": {"bar": "baz", "foo": [3, 4, 5]},
    }
    assert json.loads(str(map0)) == ref


def test_api():
    doc = Doc()
    n = 5
    keys = [f"key{i}" for i in range(n)]
    values = [f"value{i}" for i in range(n)]
    items = {keys[i]: values[i] for i in range(n)}
    map0 = Map(items)
    doc["map0"] = map0
    with pytest.raises(RuntimeError) as excinfo:
        map0[1] = 2
    assert str(excinfo.value) == "Key must be of type string"
    key_list = list(map0.keys())
    value_list = list(map0.values())
    assert len(key_list) == n
    assert len(value_list) == n
    # Yrs Map doesn't keep order
    assert {key for key in map0} == set(keys)
    assert set(key_list) == set(keys)
    assert set(value_list) == set(values)
    assert dict(map0.items()) == items
    assert "key0" in map0
    assert "key5" not in map0
    assert map0.get("key1") == "value1"
    assert map0.get("key5") is None
    assert map0.get("key5", "value5") == "value5"
    assert map0.pop("key0") == "value0"
    assert "key0" not in map0
    with pytest.raises(KeyError) as excinfo:
        del map0["key0"]
    assert str(excinfo.value) == "'key0'"
    with pytest.raises(KeyError) as excinfo:
        map0.pop("key5")
    assert map0.pop("key5", "value5") == "value5"
    with pytest.raises(RuntimeError) as excinfo:
        del map0[0]
    assert str(excinfo.value) == "Key must be of type string"
    map0.clear()
    assert len(map0) == 0

    # pop
    doc = Doc()
    map0 = Map({"foo": 1, "bar": 2})
    doc["map0"] = map0
    v = map0.pop("foo")
    assert v == 1
    assert str(map0) == '{"bar":2}'
    v = map0.pop("bar")
    assert v == 2
    assert str(map0) == "{}"

    # pop nested
    nested_doc = Doc()
    nested_doc["text"] = Text("text in subdoc")
    map0.update({"baz": nested_doc})
    v = map0.pop("baz")
    assert str(v["text"]) == "text in subdoc"
    assert str(map0) == "{}"

    nested_text = Text("abc")
    map0.update({"baz": nested_text})
    v = map0.pop("baz")
    assert v == "abc"
    assert str(map0) == "{}"

    nested_array = Array([1, 2, 3])
    map0.update({"baz": nested_array})
    v = map0.pop("baz")
    assert v == [1.0, 2.0, 3.0]
    assert str(map0) == "{}"

    nested_map = Map({"x": "y"})
    map0.update({"baz": nested_map})
    v = map0.pop("baz")
    assert v == {"x": "y"}
    assert str(map0) == "{}"


def test_to_py():
    doc = Doc()
    submap = Map({"foo": "bar"})
    subarray = Array([0, submap])
    doc["map_"] = map_ = Map({"key0": "val0", "key1": subarray})
    assert map_.to_py() == {"key0": "val0", "key1": [0, {"foo": "bar"}]}


def test_prelim():
    map0 = Map({"0": 1})
    assert map0.to_py() == {"0": 1}

    map1 = Map()
    assert map1.to_py() is None


def test_observe():
    doc = Doc()
    doc["map0"] = map0 = Map()
    doc["map1"] = map1 = Map()
    events = []

    sub = map0.observe(partial(callback, events))
    print(f"{sub=}")
    map0["0"] = 0
    assert (
        str(events[0])
        == """{target: {"0":0}, keys: {'0': {'action': 'add', 'newValue': 0.0}}, path: []}"""
    )
    events.clear()
    map0.unobserve(sub)
    map0["1"] = 1
    assert events == []

    deep_events = []
    sub = map1.observe_deep(partial(callback_deep, deep_events))
    map1["1"] = 1
    assert (
        str(deep_events[0][0])
        == """{target: {"1":1}, keys: {'1': {'action': 'add', 'newValue': 1.0}}, path: []}"""
    )
    deep_events.clear()
    map1.unobserve(sub)
    map1["0"] = 0
    assert deep_events == []



from __future__ import annotations

from typing import TYPE_CHECKING

from ._base import BaseType
from ._pycrdt import (
    StackItem,
)
from ._pycrdt import (
    UndoManager as _UndoManager,
)

if TYPE_CHECKING:  # pragma: no cover
    from ._doc import Doc


class UndoManager:
    def __init__(
        self,
        *,
        doc: Doc | None = None,
        scopes: list[BaseType] = [],
        capture_timeout_millis: int = 500,
    ) -> None:
        if doc is None:
            if not scopes:
                raise RuntimeError("UndoManager must be created with doc or scopes")
            doc = scopes[0].doc
        elif scopes:
            raise RuntimeError("UndoManager must be created with doc or scopes")
        self._undo_manager = _UndoManager(doc._doc, capture_timeout_millis)
        for scope in scopes:
            self.expand_scope(scope)

    def expand_scope(self, scope: BaseType) -> None:
        method = getattr(self._undo_manager, f"expand_scope_{scope.type_name}")
        method(scope._integrated)

    def can_undo(self) -> bool:
        return self._undo_manager.can_undo()

    def undo(self) -> bool:
        return self._undo_manager.undo()

    def can_redo(self) -> bool:
        return self._undo_manager.can_redo()

    def redo(self) -> bool:
        return self._undo_manager.redo()

    def clear(self) -> None:
        self._undo_manager.clear()

    @property
    def undo_stack(self) -> list[StackItem]:
        return self._undo_manager.undo_stack()

    @property
    def redo_stack(self) -> list[StackItem]:
        return self._undo_manager.redo_stack()



from __future__ import annotations

from types import TracebackType
from typing import TYPE_CHECKING

from ._pycrdt import Transaction as _Transaction

if TYPE_CHECKING:  # pragma: no cover
    from ._doc import Doc


class Transaction:
    _doc: Doc
    _txn: _Transaction | None
    _nb: int

    def __init__(self, doc: Doc, _txn: _Transaction | None = None) -> None:
        self._doc = doc
        self._txn = _txn
        self._nb = 0

    def __enter__(self) -> Transaction:
        self._nb += 1
        if self._txn is None:
            self._txn = self._doc._doc.create_transaction()
        self._doc._txn = self
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._nb -= 1
        # only drop the transaction when exiting root context manager
        # since nested transactions reuse the root transaction
        if self._nb == 0:
            # dropping the transaction will commit, no need to do it
            # self._txn.commit()
            assert self._txn is not None
            self._txn.drop()
            self._txn = None
            self._doc._txn = None


class ReadTransaction(Transaction):
    pass



import gc

from objsize import get_deep_size
from pycrdt import Array, Doc, Map, Text


def test_memory():
    doc = Doc()
    iter = 100
    for type_i, Type in enumerate([Array, Map, Text]):
        doc[str(type_i)] = type_ = Type()
        size0 = get_deep_size(type_)
        sizes = [size0]
        subscriptions = []
        for i in range(iter):
            subscriptions.append(type_.observe(lambda x: x))
            sizes.append(get_deep_size(type_))
            assert sizes[-1] > sizes[-2]
        sizes.pop()
        for i in range(iter):
            subscription = subscriptions.pop()
            type_.unobserve(subscription)
            size = sizes.pop()
            gc.collect()
            # this doesn't seem to always hold true:
            # assert get_deep_size(type_) == size
            assert get_deep_size(type_) >= size
        # in the end this is what really matters:
        assert get_deep_size(type_) == size0



from functools import partial

import y_py as Y
from pycrdt import Doc, Text

hello = "Hello"
world = ", World"
punct = "!"


def callback(events, event):
    events.append(
        dict(
            delta=event.delta,
            path=event.path,
        )
    )


def test_text():
    doc = Doc()
    ypy_doc = Y.YDoc()

    # add text
    state = doc.get_state()
    text = Text()
    doc["text"] = text
    with doc.transaction():
        text += hello
        with doc.transaction():
            text += world
        text += punct
    update = doc.get_update(state)
    Y.apply_update(ypy_doc, update)
    remote_text = ypy_doc.get_text("text")
    assert str(remote_text) == hello + world + punct

    # del text
    state = doc.get_state()
    with doc.transaction():
        del text[len(hello) :]
    update = doc.get_update(state)
    Y.apply_update(ypy_doc, update)
    assert str(remote_text) == hello


def test_observe():
    # pycrdt
    doc = Doc()
    text = Text()
    doc["text"] = text
    events = []

    subscription_id = text.observe(partial(callback, events))
    text += hello
    text += world
    text += punct
    del text[len(hello) : len(hello) + len(world)]
    text[len(hello) : len(hello)] = hello
    text.unobserve(subscription_id)
    text += punct

    # ypy
    ypy_doc = Y.YDoc()
    ypy_text = ypy_doc.get_text("text")
    ypy_events = []

    def ypy_callback(event):
        ypy_events.append(
            dict(
                delta=event.delta,
                path=event.path(),
            )
        )

    ypy_text.observe(ypy_callback)
    with ypy_doc.begin_transaction() as txn:
        ypy_text.extend(txn, hello)
    with ypy_doc.begin_transaction() as txn:
        ypy_text.extend(txn, world)
    with ypy_doc.begin_transaction() as txn:
        ypy_text.extend(txn, punct)
    with ypy_doc.begin_transaction() as txn:
        ypy_text.delete_range(txn, len(hello), len(world))
    with ypy_doc.begin_transaction() as txn:
        ypy_text.insert(txn, len(hello), hello)

    ref = [
        {"delta": [{"insert": hello}], "path": []},
        {"delta": [{"retain": len(hello)}, {"insert": world}], "path": []},
        {"delta": [{"retain": len(hello) + len(world)}, {"insert": punct}], "path": []},
        {"delta": [{"retain": len(hello)}, {"delete": len(world)}], "path": []},
        {"delta": [{"retain": len(hello)}, {"insert": hello}], "path": []},
    ]
    assert events == ypy_events == ref



import pytest
from pycrdt import Array, Doc, Map, Text, UndoManager


def undo_redo(data, undo_manager, val0, val1, val3):
    assert undo_manager.can_undo()
    undone = undo_manager.undo()
    assert undone
    assert data.to_py() == val1
    assert undo_manager.can_undo()
    undone = undo_manager.undo()
    assert undone
    assert data.to_py() == val0
    assert not undo_manager.can_undo()
    undone = undo_manager.undo()
    assert not undone
    assert undo_manager.can_redo()
    redone = undo_manager.redo()
    assert redone
    assert data.to_py() == val1
    assert undo_manager.can_redo()
    redone = undo_manager.redo()
    assert redone
    assert data.to_py() == val3
    assert not undo_manager.can_redo()
    redone = undo_manager.redo()
    assert not redone
    assert undo_manager.can_undo()
    undo_manager.clear()
    assert not undo_manager.can_undo()


def test_text_undo():
    doc = Doc()
    doc["data"] = data = Text()
    undo_manager = UndoManager(scopes=[data], capture_timeout_millis=0)
    val0 = ""
    val1 = "Hello"
    val2 = ", World!"
    val3 = val1 + val2
    data += val1
    assert data.to_py() == val1
    data += val2
    assert data.to_py() == val3
    undo_redo(data, undo_manager, val0, val1, val3)


def test_array_undo():
    doc = Doc()
    doc["data"] = data = Array()
    undo_manager = UndoManager(scopes=[data], capture_timeout_millis=0)
    val0 = []
    val1 = ["foo"]
    val2 = ["bar"]
    val3 = val1 + val2
    data += val1
    assert data.to_py() == val1
    data += val2
    assert data.to_py() == val3
    undo_redo(data, undo_manager, val0, val1, val3)


def test_map_undo():
    doc = Doc()
    doc["data"] = data = Map()
    undo_manager = UndoManager(scopes=[data], capture_timeout_millis=0)
    val0 = {}
    val1 = {"key0": "val0"}
    val2 = {"key1": "val1"}
    val3 = dict(**val1, **val2)
    data.update(val1)
    assert data.to_py() == val1
    data.update(val2)
    assert data.to_py() == val3
    undo_redo(data, undo_manager, val0, val1, val3)


def test_scopes():
    doc = Doc()
    doc["text"] = text = Text()
    doc["array"] = array = Array()
    doc["map"] = map = Map()
    undo_manager = UndoManager(scopes=[text], capture_timeout_millis=0)

    text += "Hello"
    text += ", World!"
    assert str(text) == "Hello, World!"
    undo_manager.undo()
    assert str(text) == "Hello"

    array.append(0)
    assert array.to_py() == [0]
    undo_manager.undo()
    assert array.to_py() == [0]
    undo_manager.expand_scope(array)
    array.append(1)
    assert array.to_py() == [0, 1]
    undo_manager.undo()
    assert array.to_py() == [0]

    map["key0"] = "val0"
    assert map.to_py() == {"key0": "val0"}
    undo_manager.undo()
    assert map.to_py() == {"key0": "val0"}
    undo_manager.expand_scope(map)
    map["key1"] = "val1"
    assert map.to_py() == {"key0": "val0", "key1": "val1"}
    undo_manager.undo()
    assert map.to_py() == {"key0": "val0"}


def test_wrong_creation():
    with pytest.raises(RuntimeError) as excinfo:
        UndoManager()
    assert str(excinfo.value) == "UndoManager must be created with doc or scopes"

    doc = Doc()
    doc["text"] = text = Text()
    with pytest.raises(RuntimeError) as excinfo:
        UndoManager(doc=doc, scopes=[text])
    assert str(excinfo.value) == "UndoManager must be created with doc or scopes"


def test_undo_redo_stacks():
    doc = Doc()
    doc["text"] = text = Text()
    undo_manager = UndoManager(scopes=[text], capture_timeout_millis=0)
    assert len(undo_manager.undo_stack) == 0
    assert len(undo_manager.redo_stack) == 0
    text += "Hello"
    assert len(undo_manager.undo_stack) == 1
    assert len(undo_manager.redo_stack) == 0
    text += ", World!"
    assert len(undo_manager.undo_stack) == 2
    assert len(undo_manager.redo_stack) == 0
    undo_manager.undo()
    assert len(undo_manager.undo_stack) == 1
    assert len(undo_manager.redo_stack) == 1
    undo_manager.undo()
    assert len(undo_manager.undo_stack) == 0
    assert len(undo_manager.redo_stack) == 2



import json
from functools import partial

import pytest
from pycrdt import Array, Doc, Map, Text


def callback(events, event):
    events.append(
        dict(
            delta=event.delta,
            path=event.path,
        )
    )


def test_str():
    doc = Doc()
    map2 = Map({"key": "val"})
    array1 = Array([2, 3, map2])
    map1 = Map({"foo": array1})
    array0 = Array([0, 1, None, map1])
    doc["array"] = array0
    assert str(array0) == '[0,1,null,{"foo":[2,3,{"key":"val"}]}]'


def test_nested():
    doc = Doc()
    text1 = Text("my_text1")
    array1 = Array([0, "foo", 2])
    text2 = Text("my_text2")
    map1 = Map({"foo": [3, 4, 5], "bar": "hello", "baz": text2})
    array0 = Array([text1, array1, map1])
    doc["array"] = array0
    ref = [
        "my_text1",
        [0, "foo", 2],
        {"bar": "hello", "foo": [3, 4, 5], "baz": "my_text2"},
    ]
    assert json.loads(str(array0)) == ref
    assert isinstance(array0[2], Map)
    assert isinstance(array0[2]["baz"], Text)


def test_array():
    doc = Doc()
    array = Array()
    doc["array"] = array
    events = []

    sub = array.observe(partial(callback, events))
    ref = [
        -1,
        -2,
        "foo",
        10,
        11,
        12,
        3.1,
        False,
        [4, 5.2],
        {"foo": 3, "bar": True, "baz": [6, 7]},
        -3,
        -4,
        -6,
        -7,
    ]
    with doc.transaction():
        array.append("foo")
        array.append(1)
        array.append(2)
        array.append(3)
        array.append(3.1)
        array.append(False)
        array.append([4, 5.2])
        array.append({"foo": 3, "bar": True, "baz": [6, 7]})
        del array[1]
        del array[1:3]
        array[1:1] = [10, 11, 12]
        array = [-1, -2] + array
        array = array + [-3, -4]
        array += [-5]
        array[-1] = -6
        array.extend([-7])

    assert json.loads(str(array)) == ref
    assert len(array) == len(ref)
    assert array[9] == ref[9]
    assert array[1:10:2] == ref[1:10:2]

    assert events == [
        {
            "delta": [{"insert": ref}],
            "path": [],
        }
    ]

    array.clear()
    assert array.to_py() == []

    events.clear()
    array.unobserve(sub)
    array.append("foo")
    assert events == []


def test_observe():
    doc = Doc()
    array = Array()
    doc["array"] = array

    sub0 = array.observe(lambda x: x)  # noqa: F841
    sub1 = array.observe(lambda x: x)  # noqa: F841
    sub2 = array.observe_deep(lambda x: x)  # noqa: F841
    sub3 = array.observe_deep(lambda x: x)  # noqa: F841

    deep_events = []

    def cb(events):
        deep_events.append(events)

    sid4 = array.observe_deep(cb)
    array.append("bar")
    assert str(deep_events[0][0]) == """{target: ["bar"], delta: [{'insert': ['bar']}], path: []}"""
    deep_events.clear()
    array.unobserve(sid4)
    array.append("baz")
    assert deep_events == []


def test_api():
    # pop
    doc = Doc()
    array = Array([1, 2, 3])
    doc["array"] = array
    v = array.pop()
    assert v == 3
    v = array.pop(0)
    assert v == 1
    assert str(array) == "[2]"

    # pop nested
    doc = Doc()
    array = Array([])
    doc["array"] = array
    nested_doc = Doc()
    nested_doc["text"] = Text("text in subdoc")
    array.insert(0, nested_doc)
    v = array.pop()
    assert str(v["text"]) == "text in subdoc"
    assert str(array) == "[]"

    nested_text = Text("abc")
    array.insert(0, nested_text)
    v = array.pop()
    assert v == "abc"
    assert str(array) == "[]"

    nested_array = Array([4, 5, 6])
    array.insert(0, nested_array)
    v = array.pop()
    assert v == [4.0, 5.0, 6.0]
    assert str(array) == "[]"

    nested_map = Map({"x": "y"})
    array.insert(0, nested_map)
    v = array.pop()
    assert v == {"x": "y"}
    assert str(array) == "[]"

    # insert
    doc = Doc()
    array = Array([1, 2, 3])
    doc["array"] = array
    array.insert(1, 4)
    assert str(array) == "[1,4,2,3]"

    # slices
    doc = Doc()
    array = Array([i for i in range(10)])
    doc["array"] = array
    with pytest.raises(RuntimeError) as excinfo:
        array[::2] = 1
    assert str(excinfo.value) == "Step not supported"
    with pytest.raises(RuntimeError) as excinfo:
        array[1:2] = 1
    assert str(excinfo.value) == "Start and stop must be equal"
    with pytest.raises(RuntimeError) as excinfo:
        array[-1:-1] = 1
    assert str(excinfo.value) == "Index out of range"
    with pytest.raises(RuntimeError) as excinfo:
        array["a"] = 1
    assert str(excinfo.value) == "Index must be of type integer"
    with pytest.raises(RuntimeError) as excinfo:
        array.pop("a")
    assert str(excinfo.value) == "Index must be of type integer"
    with pytest.raises(IndexError) as excinfo:
        array.pop(len(array))
    assert str(excinfo.value) == "Array index out of range"
    with pytest.raises(RuntimeError) as excinfo:
        del array[::2]
    assert str(excinfo.value) == "Step not supported"
    with pytest.raises(RuntimeError) as excinfo:
        del array[-1:]
    assert str(excinfo.value) == "Negative start not supported"
    with pytest.raises(RuntimeError) as excinfo:
        del array[:-1]
    assert str(excinfo.value) == "Negative stop not supported"
    with pytest.raises(TypeError) as excinfo:
        del array["a"]
    assert str(excinfo.value) == "array indices must be integers or slices, not str"

    assert [value for value in array] == [value for value in range(10)]
    assert 1 in array

    array = Array([0, 1, 2])
    assert array.to_py() == [0, 1, 2]

    array = Array()
    assert array.to_py() is None


def test_move():
    doc = Doc()
    doc["array"] = array = Array([1, 2, 3, 4])
    array.move(1, 3)
    assert str(array) == "[1,3,2,4]"


def test_to_py():
    doc = Doc()
    submap = Map({"foo": "bar"})
    subarray = Array([1, submap])
    doc["array"] = array = Array([0, subarray])
    assert array.to_py() == [0, [1, {"foo": "bar"}]]



from pycrdt import Array, Doc, Map, Text


def test_callback_transaction():
    text = Text()
    array = Array()
    map_ = Map()
    Doc(
        {
            "text": text,
            "array": array,
            "map": map_,
        }
    )
    events = []

    def callback(event):
        target = event.target
        doc = target.doc
        with doc.transaction():
            events.append(target.to_py())
            with doc.transaction():
                events.append(str(target))

    sub0 = text.observe(callback)  # noqa: F841
    sub1 = array.observe(callback)  # noqa: F841
    sub2 = map_.observe(callback)  # noqa: F841
    with text.doc.transaction():
        text += "hello"
        text += " world"
    array.append(1)
    map_["foo"] = "bar"
    assert events == [
        "hello world",
        "hello world",
        [1],
        "[1]",
        {"foo": "bar"},
        '{"foo":"bar"}',
    ]



from functools import partial

import pytest
from pycrdt import Array, Doc, Map, Text


def callback(events, event):
    events.append(event)


def encode_client_id(client_id_bytes):
    client_id_len = len(client_id_bytes)
    b = []
    for i, n in enumerate(client_id_bytes[::-1]):
        j = n << i
        if i != client_id_len - 1:
            j |= 0x80
        b.append(j)
    return bytes(b)


def test_api():
    doc = Doc()

    with pytest.raises(RuntimeError) as excinfo:
        doc[0] = Array()
    assert str(excinfo.value) == "Key must be of type string"

    doc["a0"] = a0 = Array()
    doc["m0"] = m0 = Map()
    doc["t0"] = t0 = Text()
    a1 = doc.get("a1", type=Array)
    m1 = doc.get("m1", type=Map)
    t1 = doc.get("t1", type=Text)
    assert {key for key in doc} == {"a0", "m0", "t0", "a1", "m1", "t1"}
    assert {type(value) for value in doc.values()} == {
        type(value) for value in (a0, m0, t0, a1, m1, t1)
    }
    assert {(key, type(value)) for key, value in doc.items()} == {
        (key, type(value))
        for key, value in (
            ("a0", a0),
            ("m0", m0),
            ("t0", t0),
            ("a1", a1),
            ("m1", m1),
            ("t1", t1),
        )
    }


def test_subdoc():
    doc0 = Doc()
    map0 = Map()
    doc0["map0"] = map0

    doc1 = Doc()
    map1 = Map()
    doc1["map1"] = map1

    doc2 = Doc()
    array2 = Array()
    doc2["array2"] = array2

    doc0["array0"] = Array(["hello", 1, doc1])
    map0.update({"key0": "val0", "key1": doc2})

    update0 = doc0.get_update()

    remote_doc = Doc()
    events = []
    sub = remote_doc.observe_subdocs(partial(callback, events))  # noqa: F841
    remote_doc.apply_update(update0)
    remote_array0 = Array()
    remote_map0 = Map()
    remote_doc["array0"] = remote_array0
    remote_doc["map0"] = remote_map0

    remote_doc1 = remote_array0[2]
    remote_doc2 = remote_map0["key1"]

    remote_map1 = Map()
    remote_array2 = Array()
    remote_doc1["map1"] = remote_map1
    remote_doc2["array2"] = remote_array2

    map1["foo"] = "bar"

    update1 = doc1.get_update()

    array2 += ["baz", 3]

    update2 = doc2.get_update()

    remote_doc1.apply_update(update1)
    remote_doc2.apply_update(update2)

    assert str(map1) == str(remote_map1)
    assert str(array2) == str(remote_array2)

    assert len(events) == 1
    event = events[0]
    assert len(event.added) == 2
    assert event.added[0] in (doc1.guid, doc2.guid)
    assert event.added[1] in (doc1.guid, doc2.guid)
    assert doc1.guid != doc2.guid
    assert event.removed == []
    assert event.loaded == []


def test_doc_in_event():
    doc = Doc()
    doc["array"] = array = Array()
    events = []
    sub = array.observe(partial(callback, events))  # noqa: F841
    array.append(Doc())
    assert isinstance(events[0].delta[0]["insert"][0], Doc)


def test_transaction_event():
    doc = Doc()
    events = []
    sub = doc.observe(partial(callback, events))  # noqa: F841
    doc["text0"] = Text("Hello, World!")

    remote_doc = Doc()
    for event in events:
        remote_doc.apply_update(event.update)
    events.clear()

    doc.unobserve(sub)
    doc["text1"] = Text("Goodbye!")
    assert len(events) == 0

    remote_text0 = remote_doc.get("text0", type=Text)
    assert str(remote_text0) == "Hello, World!"
    remote_text1 = remote_doc.get("text1", type=Text)
    assert str(remote_text1) == ""


def test_client_id():
    doc0 = Doc()
    doc1 = Doc()
    assert doc0.client_id != doc1.client_id

    client_id_bytes = b"\x01\x02\x03\x04"
    client_id = int.from_bytes(client_id_bytes, byteorder="big")
    doc2 = Doc(client_id=client_id)
    assert doc2.client_id == client_id
    text = Text("Hello, World!")
    doc2["text"] = text
    update = doc2.get_update()

    b = encode_client_id(client_id_bytes)

    assert update[2 : 2 + len(b)] == b


def test_roots():
    remote_doc = Doc(
        {
            "a": Text("foo"),
            "b": Array([5, 2, 8]),
            "c": Map({"k1": 1, "k2": 2}),
        }
    )
    roots = dict(remote_doc)
    assert str(roots["a"]) == "foo"
    assert list(roots["b"]) == [5, 2, 8]
    assert dict(roots["c"]) == {"k1": 1, "k2": 2}

    local_doc = Doc()
    update = remote_doc.get_update()
    local_doc.apply_update(update)
    roots = dict(local_doc)
    assert roots["a"] is None
    assert roots["b"] is None
    assert roots["c"] is None
    # assert str(roots["a"]) == "foo"
    # assert list(roots["b"]) == [5, 2, 8]
    # assert dict(roots["c"]) == None  # {"k1": 1, "k2": 2}


def test_empty_update():
    doc = Doc()
    doc["text"] = Text()
    events = []
    sub = doc.observe(partial(callback, events))  # noqa: F841

    # this triggers an empty update
    doc["text"]
    # empty updates should not emit an event
    assert not events


def test_not_empty_update():
    doc = Doc()
    doc["text"] = text = Text()
    events = []
    sub = doc.observe(partial(callback, events))  # noqa: F841

    text += "helloo"
    events.clear()
    del text[5]
    assert events


def test_get_update_exception():
    doc = Doc()
    with pytest.raises(ValueError) as excinfo:
        doc.get_update(b"\x12")
    assert str(excinfo.value) == "Cannot decode state"
